/*
 *  $Id: ElementMap.java,v 1.9 2001/11/28 20:36:20 fekete Exp $
 *  Copyright (C) 1998 Jean-Daniel Fekete and Ecole des Mines de Nantes
 *
 *  This file is part of Compus.
 *  Compus is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import java.awt.Color;
import java.util.Arrays;
import java.util.*;
import javax.swing.ListModel;
import javax.swing.event.ListDataListener;
import javax.swing.event.EventListenerList;
import javax.swing.event.ListDataEvent;

public class ElementMap implements ListModel {
  HashMap set;
  String keys_table[];
  protected EventListenerList listenerList = new EventListenerList();
  int last_color;
  volatile ElementInfo cached_info;
  volatile String cached_key;

  static class ElementInfo implements Cloneable {
    String name;
    IndirectColor icolor;
    int surface;
    int occurences;
    float meanSurface;
    ElementType type;

    ElementInfo(String name, Color color, int surface) {
      this.name = name;
      this.icolor = new IndirectColor(color);
      this.surface = surface;
      occurences = 1;
      meanSurface = 0;
    }
    public void addValue(String text) {
      ElementType t = new ElementType(text);
      if (type == null) {
	type = t;
      }
      else {
	type.mergeTypes(t);
      }
    }
    public void addInfo(ElementInfo info) {
      info.icolor = icolor;	// share the icolor
      surface += info.surface;
      occurences += info.occurences;
      meanSurface = 0;
      if (type != null)
	type.mergeTypes(info.type);
    }
    public final float computeMean() {
      if (meanSurface == 0) {
	meanSurface = ((float)surface) / occurences;
	if (type != null)
	  meanSurface -= 2*name.length() + 5;
      }
      return meanSurface;
    }
    public Color getColor() { return icolor.getColor(); }
    public void setColor(Color c) { icolor.setColor(c); }

    public boolean isVisible() { return icolor.getColor().getAlpha() != 0; }
    public void setVisible(boolean t) {
      int alpha = t ? 255 : 0;
      Color c = icolor.getColor();
      if (c.getAlpha() != alpha) {
	icolor.setColor(new Color(c.getRed(), c.getGreen(), c.getBlue(),
				  alpha));
      }
    }

    public ElementType getType() { return type; }
    public void setType(ElementType t) { type = t; }

    public final String toString() {
      return name+
	", "+getColor()+
	", s="+surface+
	", o="+occurences+
	", m="+meanSurface+
	", t="+type;
    }
    public Object clone() {
      try {
	return super.clone();
      } catch(CloneNotSupportedException e) {
	return null;
      }
    }
  }

  public ElementMap(int initialCapacity, float loadFactor) {
    set = new HashMap(initialCapacity, loadFactor);
  }
  public ElementMap(int initialCapacity) {
    set = new HashMap(initialCapacity);
  }
  public ElementMap() {
    set = new HashMap();
  }
  
  public ElementInfo getInfo(String key) {
    if (cached_key != key) {
      cached_key = key;
      cached_info = (ElementInfo)set.get(key);
    }
    return cached_info;
  }

  public  Color put(String key, Color color, int size) {
    ElementInfo info = getInfo(key);
    if (info == null) {
      cached_info = new ElementInfo(key, color, size);
      set.put(key, cached_info);
      keys_table = null;
    }
    else {
      Color c = info.getColor();
      info.setColor(color);
      info.surface += size;
      info.occurences++;
      return c;
    }
    return color;
  }
  public void addElement(ElementInfo ninfo) {
    ElementInfo info = getInfo(ninfo.name);
    //System.out.println("Adding info from "+ninfo.name);

    if (info == null) {
      cached_info = (ElementInfo)ninfo.clone();
      set.put(ninfo.name, cached_info);
      keys_table = null;
    }
    else {
      info.addInfo(ninfo);
    }
  }
  
  public Color putColor(String key, Color color) {
    ElementInfo info = getInfo(key);
    if (info == null) {
      throw new NoSuchElementException("Color change in ElementMap");
    }
    else {
      Color c = info.getColor();
      info.setColor(color);
      return c;
    }
  }
  public void addElements(ElementMap map) {
    Iterator iter = map.set.values().iterator();
    while (iter.hasNext()) {
      addElement((ElementInfo)iter.next());
    }
  }
  public void updateContent(String key, String content) {
    ElementInfo info = getInfo(key);
    if (info == null) {
      throw new NoSuchElementException("Content in ElementMap");
    }
    else {
      info.addValue(content);
    }
  }
  public ElementType getType(String key) {
    ElementInfo info = getInfo(key);
    if (info == null) {
      throw new NoSuchElementException("Content in ElementMap");
    }
    else {
      return info.type;
    }
  }
  public boolean isVisible(String key) {
    ElementInfo info = getInfo(key);
    if (info == null) {
      throw new NoSuchElementException("Content in ElementMap");
    }
    else {
      return info.isVisible();
    }
  }
  public void setVisible(String key, boolean t) {
    ElementInfo info = getInfo(key);
    if (info == null) {
      throw new NoSuchElementException("Content in ElementMap");
    }
    else {
      info.setVisible(t);
    }
  }
  public  String[] getKeys() {
    if (keys_table == null) {
      keys_table = new String[set.size()];
      Iterator iter = set.keySet().iterator();
      int i = 0;
      while (iter.hasNext()) {
	keys_table[i++] = (String)iter.next();
      }
      Arrays.sort(keys_table);
    }
    return keys_table;
  }
  public void swapKeys(int from, int to) {
    String s[] = getKeys();
    if (from != to &&
	from >= 0 && from < s.length &&
	to >= 0 && to < s.length) {
      String f = s[from];
      s[from] = s[to];
      s[to] = f;
      if (from < to) {
	fireContentsChanged(this, from, to);
      }
      else {
	fireContentsChanged(this, to, from);
      }
    }
  }
  public  void putNew(String key, int size) {
    ElementInfo info = getInfo(key);
    if (info == null) {
      put(key, newColor(), size);
    } else {
      info.surface += size;
      info.occurences++;
    }
  }
  public Color getColor(String key) {
    ElementInfo ret = getInfo(key);
    if (ret == null)
      return null;
    else
      return ret.getColor();
  }
  public int getOccurences(String key) {
    ElementInfo ret = getInfo(key);
    if (ret == null)
      return 0;
    else
      return ret.occurences;
  }
  public int getSurface(String key) {
    ElementInfo ret = getInfo(key);
    if (ret == null)
      return 0;
    else
      return ret.surface;
  }
  public int getSize() { return set.size(); }
  public Object getElementAt(int index) {
    String k[] = getKeys();
    return k[index];
  }
  public void clear() {
    int s = getSize();
    keys_table = null;
    set = new HashMap();
    fireIntervalRemoved(this, 0, s);
  }
  public void addListDataListener(ListDataListener l) {
    listenerList.add(ListDataListener.class, l);
  }

  public void removeListDataListener(ListDataListener l) {
    listenerList.remove(ListDataListener.class, l);
  }
  public void fireContentsChanged(Object source, int index0, int index1)
  {
    Object[] listeners = listenerList.getListenerList();
    ListDataEvent e = null;

    for (int i = listeners.length - 2; i >= 0; i -= 2) {
      if (listeners[i] == ListDataListener.class) {
	if (e == null) {
	  e = new ListDataEvent(source, ListDataEvent.CONTENTS_CHANGED, index0, index1);
	}
	((ListDataListener)listeners[i+1]).contentsChanged(e);
      }	       
    }
  }
  protected void fireIntervalRemoved(Object source, int index0, int index1)
  {
    Object[] listeners = listenerList.getListenerList();
    ListDataEvent e = null;

    for (int i = listeners.length - 2; i >= 0; i -= 2) {
      if (listeners[i] == ListDataListener.class) {
	if (e == null) {
	  e = new ListDataEvent(source, ListDataEvent.INTERVAL_REMOVED, index0, index1);
	}
	((ListDataListener)listeners[i+1]).intervalRemoved(e);
      }	       
    }
  }

  public Color newColor() {
    if (last_color < color_table.length)
      return color_table[last_color++];
    else
      return color_table[last_color++ % color_table.length]; // +++ fix
  }

  static final int alpha = 255;
  static Color color_table[] = {
    Color.black,
    new Color(255, 255, 0, alpha), // Color.yellow, 
    new Color(192, 192, 192, alpha), // Color.lightGray,
    new Color(64, 64, 64, alpha), // Color.darkGray,
    new Color(128, 128, 128, alpha), // Color.gray,
    new Color(255, 175, 175, alpha), // Color.pink,
    new Color(255, 200, 0, alpha), //Color.orange,
    new Color(255, 0, 255, alpha), // Color.magenta,
    new Color(0, 255, 255, alpha), // Color.cyan,
    new Color(255, 0, 0, alpha), // Color.red
    new Color(0, 255, 0, alpha), // Color.green,
    new Color(0, 0, 255, alpha), // Color.blue,
  };
  static Color vivid_colors[] = {
    Color.red,
    Color.green,
    Color.blue,
    Color.cyan,
    Color.magenta,
  };
  static Color medium_colors[] = {
    new Color(0x66, 0x88, 0x00),
    new Color(0xff, 0x44, 0x55),
    new Color(0x11, 0xaa, 0x22),
    new Color(0x00, 0x99, 0x88),
    new Color(0xdd, 0x66, 0x11),
    new Color(0x77, 0x66, 0xcc),
    new Color(0xcc, 0x66, 0x00),
    /*
    Color.orange,
    Color.pink,
    new Color(128, 0, 0),
    new Color(0, 128, 0),
    new Color(0, 0, 128),
    */
  };
  static Color light_colors[] = {
    Color.white,
    Color.gray,
    Color.darkGray,
    Color.lightGray,
    Color.yellow
  };
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("ElementMap:\n");
    Object keys[] = getKeys();
    for (int i = 0; i < keys.length; i++) {
      sb.append('\t').append(keys[i]).append(": ").append(set.get(keys[i])).append('\n');
    }
    return sb.toString();
  }
  static final int sgn(float f1, float f2) {
    float sgn = f1 - f2;
    if (sgn < 0) return -1;
    else if (sgn > 0) return 1;
    else return 0;
  }

  public void optimize() {
    ElementInfo table[] = new ElementInfo[set.size()];
    Iterator iter = set.values().iterator();
    int i = 0;
    while (iter.hasNext()) {
      table[i] = (ElementInfo)iter.next();
      if (table[i].getColor().getAlpha() != 0)
	i++;
    }
    int end = i;
    Arrays.sort(table, 0, end, new Comparator() {
      public int compare(Object o1, Object o2) {
	float v1 = ((ElementInfo)o1).computeMean();
	float v2 = ((ElementInfo)o2).computeMean();
	return sgn(v1, v2);
      }
    });
    if (end <= vivid_colors.length) {
      for (i = 0; i < end; i++) {
	table[i].setColor(vivid_colors[i % vivid_colors.length]);
      }
    } else {
      for (i = 0; i < end && table[i].computeMean() < 20.0f; i++) {
	table[i].setColor(vivid_colors[i % vivid_colors.length]);
      }
      int first = i;
      for (; i < end && table[i].computeMean() < 100.0f; i++) {
	table[i].setColor(medium_colors[(i - first) % medium_colors.length]);
      }
      first = end - 1;	// so that the last gets the lightest color
      for (; i < end; i++) {
	table[i].setColor(light_colors[(first - i) % light_colors.length]);
      }
    }
    fireContentsChanged(this, 0, table.length);
  }
}
