/*
 *  $Id: ElementType.java,v 1.5 2001/11/28 20:36:20 fekete Exp $
 *  Copyright (C) 1998 Jean-Daniel Fekete and Ecole des Mines de Nantes
 *
 *  This file is part of Compus.
 *  Compus is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import java.text.*;
import java.util.Date;
import java.util.Calendar;

public class ElementType {

  static public final int TYPE_EMPTY = 0;
  static public final int TYPE_INT = 1;
  static public final int TYPE_REAL = 2;
  static public final int TYPE_DATE = 3;
  static public final int TYPE_TEXT = 4;
  static public final String type_name[] = {
    "empty",
    "int",
    "real",
    "date",
    "text",
  };
  static public final char type_char[] = {
    '/',
    'I',
    'R',
    'D',
    'T',
  };

  int type;
  Date date;
  Number num;
  String text;
  double min, max;

  static DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT);
  static NumberFormat nf = NumberFormat.getInstance();

  ElementType(String text) {
    this(text, new ParsePosition(0));
  }
  ElementType(String text, int index) {
    this(text, new ParsePosition(index));
  }
  ElementType(String text, ParsePosition pos) {
    this.text = text;
    int len = text.length();
    int p = pos.getIndex();
    while (p < len &&
	   Character.isWhitespace(text.charAt(p)))
      p++;
    if (p == len) {
      type = TYPE_EMPTY;
      text = "";
      return;
    }
    pos.setIndex(p);
    // Then try for a number
    {
      Number n = nf.parse(text, pos);
      if (n != null) {
	num = n;
	min = max = doubleValue();
	if (n instanceof Double) {
	  type = TYPE_REAL;
	  text = num.toString();
	  //System.out.println(this);
	} else {
	  type = TYPE_INT;
	}
      }
    }
    // Try for a date
    p = pos.getIndex();
    if (type == TYPE_INT &&
	intValue() > 0 &&
	intValue() <= 31 &&
	p < len &&
	text.charAt(p) == '/') {
      //Date d = df.parse(text, pos);
      int day = intValue();
      Date d = null;

      pos.setIndex(p+1);
      Number n = nf.parse(text, pos);
      if (n == null) {
	//System.out.println(this);
	return;
      }
      int month = n.intValue();
      int year;
      Calendar cal = Calendar.getInstance();
      cal.clear();
      p = pos.getIndex();
      if (month < 1) {
	//System.out.println(this);
	return;
      }
      else if (month > 12) {
	year = month;
	month = day;
	cal.set(year, month - 1, 1);
	d = new Date(cal.get(Calendar.DATE));
      }
      else if (p < len &&
	       text.charAt(p) == '/') {
	pos.setIndex(p+1);
	n = nf.parse(text, pos);
	if (n == null) {
	  System.out.println(this);
	  return;
	}
	year = n.intValue();
	cal.set(year, month - 1, day);
	d = new Date(cal.get(Calendar.DATE));
      }
      if (d != null) {
	type = TYPE_DATE;
	num = null;
	date = d;
	min = max = doubleValue();
	text = df.format(d);
	//System.out.println(this);
	return;
      }
    }
    else if (type == TYPE_INT) {
      text = num.toString();
    }
    else {
      type = TYPE_TEXT;
    }
  }
  public int getType() { return type; }
  public int intValue() {
    if (num != null)
      return num.intValue();
    else
      return 0;
  }
  public long longValue() {
    if (num != null)
      return num.longValue();
    else if (date != null)
      return date.getTime();
    else
      return 0;
  }
  public float floatValue() {
    if (num != null)
      return num.floatValue();
    else if (date != null)
      return date.getTime();
    else
      return 0;
  }
  public double doubleValue() {
    if (num != null)
      return num.doubleValue();
    else if (date != null)
      return date.getTime();
    else
      return 0;
  }
  public Date dateValue() {
    return date;
  }
  public String textValue() {
    return text;
  }
  public double getMinValue() { return min; }
  public double getMaxValue() { return max; }

  public void mergeTypes(ElementType t) {
    if (t == null) return;
    int new_t = sup(t.type, type);
    type = new_t;
    if (new_t == TYPE_EMPTY || new_t == TYPE_TEXT) {
      min = max = 0;
      num = null;
      date = null;
      return;
    }
    min = Math.min(min, t.min);
    max = Math.max(max, t.max);
  }
  static public int sup(int t1, int t2) {
    if (t1 == t2) return t1;
    if (t1 < t2) {
      int t = t1;
      t1 = t2;
      t2 = t;
    }
    if (t2 == TYPE_EMPTY)
      return t1;
    switch(t1) {
    case TYPE_TEXT:
    case TYPE_DATE:
      return TYPE_TEXT;
    case TYPE_REAL:
      return TYPE_REAL;
    }
    throw new IllegalArgumentException("Unknown type value "+t1);
  }
  public String getTypeName() {
    return type_name[type];
  }
  public char getTypeChar() {
    return type_char[type];
  }
  public String toString() {
    return
      "[ElementType: "+type_name[type] +
      ": '"+textValue()+
      "' min: "+getMinValue()+
      " max: "+getMaxValue()+
      "]";
  }
}
