/*
 *  $Id: XmlDocument.java,v 1.5 2001/11/28 20:36:20 fekete Exp $
 *  Copyright (C) 1998 Jean-Daniel Fekete and Ecole des Mines de Nantes
 *
 *  This file is part of Compus.
 *  Compus is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.helpers.AttributeListImpl;
import java.util.Vector;
import java.io.*;
import java.net.URL;

public class XmlDocument
  implements EntityResolver, DTDHandler, DocumentHandler, ErrorHandler
{
  String systemId;
  String publicId;
  Vector elements;
  Vector opened;
  Vector events;
  String text;
  StringBuffer buffer;
  boolean loaded = false;
  Element current;
  boolean keep_attributes = false;

  public class Element {
    String name;
    AttributeList atts;
    int start, end;
    Vector children;
    Element parent = null;

    public Element(Element p, String n, AttributeList a, int pos) {
      parent = p;
      name = n.intern();
      if (keep_attributes) {
	atts = new AttributeListImpl(a);
      } else {
	atts = null;
      }
      start = end = pos;
    }
    public String getContents() {
      return buffer.substring(start, end);
    }
    public XmlDocument getDocument() { return XmlDocument.this; }
    public Element getParent() { return parent; }
    public void setParent(Element p) { parent = p; }
    public String getName() { return name; }
    public int size() { return end - start; }
    public boolean isEmpty() { return end == start; }
    public void addChild(Element e) {
      if (children == null)
	children = new Vector();
      children.addElement(e);
    }
    public Vector getChildren() { return children; }
    public Element getChild(int i) {
      return (Element)children.elementAt(i);
    }
    public int getChildCount() {
      return children == null ? 0 : children.size();
    }
    public boolean contains(int index) {
      return index >= start && index <= end;
    }
  }
  static public class Event {
    int type;
    Element element;

    static final int START = 0;
    static final int END = 1;

    Event(int type, Element e) {
      this.type = type;
      element = e;
    }

    int getPosition() {
      return type == START ? element.start : element.end;
    }
    Element getElement() { return element; }
  }

  public XmlDocument() {
    opened = new Vector();
    elements = new Vector();
    events = new Vector();
  }
  public XmlDocument(String name) {
    systemId = name;
    try {
      load();
    } catch(Exception e) { }
  }
  public XmlDocument(String name, Parser parser, InputSource source) {
    systemId = name;
    try {
      load(parser, source);
    } catch(Exception e) { }
  }

  public int size() {
    return (text == null) ? 0 : text.length();
  }

  public String getName() { return systemId; }
  public void setName(String name) { systemId = name; }

  public boolean isKeepingAttributes() { return keep_attributes; }
  public void setKeepingAttributes(boolean b) {
    keep_attributes = b;
  }

  public Element getElement(int i) {
    return (Element)elements.elementAt(i);
  }
  public int elementCount() { return elements.size(); }
  public int eventCount() { return events.size(); }

  public Event getEvent(int i) { return (Event)events.elementAt(i); }

  public String substring(int start, int end) {
    return text.substring(start, end);
  }
  public String getText() { return text; }

  Element getCurrentParent() {
    if (opened.size() == 0) return null;
    return (Element)opened.lastElement();
  }
  void addElement(String name, AttributeList atts) {
    Element p = getCurrentParent();
    Element e = new Element(p, name, atts, buffer.length());
    events.addElement(new Event(Event.START, e));
    elements.addElement(e);
    opened.addElement(e);
    if (p != null) 
      p.addChild(e);
  }
  void setElementEnd() {
    Element e = (Element)opened.lastElement();
    opened.removeElementAt(opened.size()-1);
    e.end = buffer.length();
    if (e.children != null) {
      e.children.trimToSize();
    }
    events.addElement(new Event(Event.END, e));
  }

  public Element rootElement() {
    if (elementCount() > 0)
      return getElement(0);
    return null;
  }

  Element moreSpecific(Element e, int index) {
    if (! e.contains(index)) return null;
    Vector c = e.getChildren();
    if (c == null)
      return e;
    for (int i = 0; i < c.size(); i++) {
      Element ce = moreSpecific((Element)c.elementAt(i), index);
      if (ce != null)
	return ce;
    }
    return e;
  }

  public Element findElementAt(int index) {
    return moreSpecific(rootElement(), index);
  }

  public InputSource resolveEntity (String publicId, String systemId)
  {
    return null;
  }
  public void notationDecl (String name, String publicId, String systemId)
  {
  }
  public void unparsedEntityDecl (String name,
				  String publicId,
				  String systemId,
				  String notationName)
  {
  }
  public void setDocumentLocator (Locator locator)
  {
//      if (systemId == null && publicId == null) {
//        systemId = locator.getPublicId();
//        publicId = locator.getSystemId();
//      }
  }
  public void startDocument ()
  {
    System.out.println("Loading "+systemId);
    buffer = new StringBuffer();
    current = null;
  }
  public void endDocument ()
  {
    //System.out.println("Finished "+systemId+", "+publicId);
    text = buffer.toString();
    buffer = null;
    current = null;
    loaded = true;
    opened = null;
    elements.trimToSize();
    events.trimToSize();
  }
  public void startElement (String name, AttributeList attributes)
  {
    addElement(name, attributes);
    buffer.append('<').append(name).append('>');
  }
  public void endElement (String name)
  {
    buffer.append('<').append('/').append(name).append('>');
    setElementEnd();
  }
  public void characters (char ch[], int start, int length)
  {
    buffer.append(ch, start, length);
  }
  public void ignorableWhitespace (char ch[], int start, int length)
  {
    //buffer.append(ch, start, length);
  }
  public void processingInstruction (String target, String data)
  {
  }
  public void warning (SAXParseException exception)
  {
    System.out.println("Warning: " +
		       exception.getMessage() +
		       " (" +
		       exception.getSystemId() +
		       ':' +
		       exception.getLineNumber() +
		       ',' +
		       exception.getColumnNumber() +
		       ')');
  }
  public void error (SAXParseException exception)
  {
    System.out.println("Recoverable Error: " +
		       exception.getMessage() +
		       " (" +
		       exception.getSystemId() +
		       ':' +
		       exception.getLineNumber() +
		       ',' +
		       exception.getColumnNumber() +
		       ')');
  }
  public void fatalError (SAXParseException exception)
  {
    System.out.println("Fatal Error: " +
		       exception.getMessage() +
		       " (" +
		       exception.getSystemId() +
		       ':' +
		       exception.getLineNumber() +
		       ',' +
		       exception.getColumnNumber() +
		       ')');
  }
  public void print(PrintStream out) {
    out.println("Document: "+systemId);
    out.println(" # of elements: "+elements.size());
    for (int i = 0; i < elements.size(); i++) {
      Element e = getElement(i);
      out.println("Name: "+e.name+" start="+e.start+" end="+e.end);
      //out.println("\t"+e.beginning(20));
    }
  }
  public synchronized void load() 
    throws Exception {
    Parser parser;

    parser = createParser();

    load(parser, fileInputSource(systemId));
  }
  public synchronized void load(Parser parser, InputSource source)
    throws Exception {
    loaded = false;
    opened = new Vector();
    elements = new Vector();
    events = new Vector();
    parser.setDTDHandler(this);
    parser.setDocumentHandler(this);
    parser.setErrorHandler(this);
    parser.parse(source);
  }

  public boolean isLoaded() { return loaded; }

  static public Parser createParser() {
    String parserClass = System.getProperty("com.jclark.xsl.sax.parser");
    if (parserClass == null)
      parserClass = System.getProperty("org.xml.sax.parser");
    if (parserClass == null)
      parserClass = "com.jclark.xml.sax.CommentDriver";
    try {
      return (Parser)Class.forName(parserClass).newInstance();
    }
    catch (ClassNotFoundException e) {
      System.err.println(e.toString());
    }
    catch (InstantiationException e) {
      System.err.println(e.toString());
    }
    catch (IllegalAccessException e) {
      System.err.println(e.toString());
    }
    catch (ClassCastException e) {
      System.err.println(parserClass + " is not a SAX driver");
    }
    return null;
  }

  static public InputSource fileInputSource(String str) {
    return fileInputSource(new File(str));
  }

  static public InputSource fileInputSource(File file) {
    String path = file.getAbsolutePath();
    String fSep = System.getProperty("file.separator");
    if (fSep != null && fSep.length() == 1)
      path = path.replace(fSep.charAt(0), '/');
    if (path.length() > 0 && path.charAt(0) != '/')
      path = '/' + path;
    try {
      return new InputSource(new URL("file", "", path).toString());
    }
    catch (java.net.MalformedURLException e) {
      /* According to the spec this could only happen if the file
	 protocol were not recognized. */
      throw new Error("unexpected MalformedURLException");
    }
  }
}
