/*
 *  $Id: XmlViewBase.java,v 1.9 2002/10/16 08:43:34 fekete Exp $
 *  Copyright (C) 1998 Jean-Daniel Fekete and Ecole des Mines de Nantes
 *
 *  This file is part of Compus.
 *  Compus is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import java.awt.*;
import javax.swing.*;
import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;
import java.util.*;
import java.io.*;
import java.util.Locale;
import java.io.IOException;
import java.net.URL;
import gnu.regexp.*;


public abstract class XmlViewBase extends JComponent {
  public static final int DISPLAY_DOC_ORDER = 0;
  public static final int DISPLAY_SORTED_ORDER = 1;
  protected int display_type = DISPLAY_DOC_ORDER;
  protected ElementMap elementMap;
  protected Image offscreen;
  protected int maxSize;
  protected String text;
  protected long surface = -1;
  protected int elements = -1;
  protected float scale;
  protected Dimension size = new Dimension();
  protected int current_start, current_end;
  protected REMatch [] matches;

  public XmlViewBase(ElementMap map) {
    elementMap = map;
    setDoubleBuffered(false);
  }
  public XmlViewBase() {
    this(null);
  }
  public String getText() { return text; }
  public int getDisplayType() { return display_type; }
  public void setDisplayType(int t) {
    if (display_type != t) {
      display_type = t;
      documentModified();	// optimize, can save surface and elements 
    }
  }
  abstract void updateElementMap();
  abstract public String getDocumentName();

  public void setMaxSize(int m) { maxSize = m; }
  public int getMaxSize() { return maxSize; }
  public int getDisplaySize() {
    getSize(size);
    return (size.width - 1) * size.height;
  }
  public ElementMap getElementMap() { return elementMap; }
  //public int getDocumentSize() { return sdoc.getLength(); }
  public int getDocumentSize() {
    switch(display_type) {
    case DISPLAY_DOC_ORDER: return text.length();
    case DISPLAY_SORTED_ORDER: return (int)getVisibleSurface();
    }
    return text.length();
  }

  public void match(String regexp) {
    matches = null;
    try {
      RE exp = new RE(regexp);
      matches = exp.getAllMatches(text);
    }
    catch(REException e) {
    }
    documentModified();
  }
  public void paint(Graphics og) {
    getSize(size);
    if (offscreen == null ||
	offscreen.getWidth(null) != size.width ||
	offscreen.getHeight(null) != size.height) {
      offscreen = createImage(size.width, size.height);
      Graphics g = offscreen.getGraphics();
      //System.out.print("begin paint...");
      updateScale();
      g.setColor(getBackground());
      g.fillRect(0, 0, size.width, size.height);
      g.setColor(Color.black);
      g.drawLine(size.width-1, 0, size.width-1, size.height);
      g.setColor(getBackground());
      current_start = current_end = 0;
      switch (display_type) {
      case DISPLAY_DOC_ORDER: paintElements(g); break;
      case DISPLAY_SORTED_ORDER: paintSorted(g); break;
      }
      addPaintRange(g, current_end, getMaxSize(), getBackground());
      paintMatches(g);
      //System.out.println("end paint");
    }
    og.drawImage(offscreen, 0, 0, null);
  }

  abstract protected void paintElements(Graphics g);

  protected void paintSorted(Graphics g) {
    ElementMap map = getElementMap();
    int start = 0;
    for (int i = 0; i < map.getSize(); i++) {
      String s = (String)map.getElementAt(i);
      if (map.getColor(s).getAlpha() != 0) {
	int size = map.getSurface(s);
	addPaintRange(g, start, start+size, map.getColor(s));
	start += size;
      }
    }
  }

  protected void paintMatches(Graphics g) {
    if (matches == null)
      return;
    Color c = getElementMap().getColor("MATCH");
    if (c == null)
      c = Color.red;
    for (int i = 0; i < matches.length; i++) {
      REMatch m = matches[i];
      paintRange(g, m.getStartIndex(), m.getEndIndex(), c);
    }
  }

  public String getSortedContext(int index) {
    ElementMap map = getElementMap();
    int start = 0;
    for (int i = 0; i < map.getSize(); i++) {
      String s = (String)map.getElementAt(i);
      if (map.getColor(s).getAlpha() != 0) {
	int size = map.getSurface(s);
	start += size;
	if (index < start)
	  return getDocumentName() + " "+s;
      }
    }
    return getDocumentName() + index;
  }

  void addPaintRange(Graphics g, int start, int end, Color c) {
    if (c == g.getColor()) {
      current_end = end;
    } else {
      paintRange(g, current_start, current_end, g.getColor());
      g.setColor(c);
      current_start = start;
      current_end = end;
    }
  }

  void paintRange(Graphics g, int start, int end, Color c) {
    start = (int)(start * scale);
    end = (int)(end * scale);
    if (start == end) return;

    g.setColor(c);
    //System.out.println("paintRange: "+start+","+end+" with "+c);
    getSize(size);
    size.width--;
    if (size.width > size.height) {	// horizontal
      //System.out.println("horizontal");
      int start_x = start / size.height;
      int start_y = start % size.height;
      int end_x = end / size.height;
      int end_y = end % size.height;

      if (start_x == end_x) {
	g.drawLine(start_x, start_y, end_x, end_y-1);
      } else {
	g.drawLine(start_x, start_y, start_x, size.height);
	g.fillRect(start_x+1, 0, end_x-start_x-1, size.height);
	g.drawLine(end_x, 0, end_x, end_y);
      }
    } else {
      //System.out.println("vertical");
      int start_x = start % size.width;
      int start_y = start / size.width;
      int end_x = end % size.width;
      int end_y = end / size.width;

      if (start_y == end_y) {
	g.drawLine(start_x, start_y, end_x-1, end_y);
      } else {
	g.drawLine(start_x, start_y, size.width-1, start_y);
	g.fillRect(0, start_y+1, size.width, end_y-start_y-1);
	g.drawLine(0, end_y, end_x-1, end_y);
      }
    }
  }
  protected void updateScale() {
    if (getMaxSize() != 0)
      scale = (float)getDisplaySize() / (float)getMaxSize();
    else
      scale = (float)getDisplaySize() / (float)getDocumentSize();
  }

  public synchronized void documentModified() {
    offscreen = null;
    repaint();
    surface = -1;
    elements = -1;
    setToolTipText(null);
  }

  public long getVisibleSurface() {
    if (surface == -1) {
      surface = 0;
      ElementMap map = getElementMap();
      for (int i = 0; i < map.getSize(); i++) {
	String s = (String)map.getElementAt(i);
	if (map.getColor(s).getAlpha() != 0) {
	  surface += map.getSurface(s);
	}
      }
    }
    return surface;
  }
  public int getVisibleElements() {
    if (elements == -1) {
      elements = 0;
      ElementMap map = getElementMap();
      String[] strs = map.getKeys();
      for (int i = 0; i < strs.length; i++) {
	String s = strs[i];
	if (map.isVisible(s)) {
	  elements += map.getOccurences(s);
	}
      }
    }
    return elements;
  }

  public int indexOf(int x, int y) {
    updateScale();
    getSize(size);
    size.width--;
    if (size.width > size.height) {
      return (int) ((x * size.height + y) / scale);
    } else {
      return (int) ((y * size.width + x) / scale);
    }
  }

  abstract public String getContext(int index);

  public boolean contains(int x, int y) {
    switch(display_type) {
    case DISPLAY_DOC_ORDER:
      setToolTipText(getContext(indexOf(x, y)));
      break;
    case DISPLAY_SORTED_ORDER:
      setToolTipText(getSortedContext(indexOf(x, y)));
      break;
    }
    return super.contains(x, y);
  }

  static public Parser createParser() {
    String parserClass = System.getProperty("com.jclark.xsl.sax.parser");
    if (parserClass == null)
      parserClass = System.getProperty("org.xml.sax.parser");
    if (parserClass == null)
      parserClass = "com.jclark.xml.sax.CommentDriver";
    try {
      return (Parser)Class.forName(parserClass).newInstance();
    }
    catch (ClassNotFoundException e) {
      System.err.println(e.toString());
    }
    catch (InstantiationException e) {
      System.err.println(e.toString());
    }
    catch (IllegalAccessException e) {
      System.err.println(e.toString());
    }
    catch (ClassCastException e) {
      System.err.println(parserClass + " is not a SAX driver");
    }
    return null;
  }

  static public InputSource fileInputSource(String str) {
    return fileInputSource(new File(str));
  }

  static public InputSource fileInputSource(File file) {
    String path = file.getAbsolutePath();
    String fSep = System.getProperty("file.separator");
    if (fSep != null && fSep.length() == 1)
      path = path.replace(fSep.charAt(0), '/');
    if (path.length() > 0 && path.charAt(0) != '/')
      path = '/' + path;
    try {
      return new InputSource(new URL("file", "", path).toString());
    }
    catch (java.net.MalformedURLException e) {
      /* According to the spec this could only happen if the file
	 protocol were not recognized. */
      throw new Error("unexpected MalformedURLException");
    }
  }

  public int getSaveableElements() {
    int elements = 0;
    ElementMap map = getElementMap();
    String strs[] = map.getKeys();

    for (int i = 0; i < strs.length; i++) {
      String s = strs[i];
      ElementMap.ElementInfo info = map.getInfo(s);
      if (info.isVisible()) {
	elements ++;
      }
    }
    return elements;
  }

  public void addSaveableElement(HashMap view, String name, String s) {
    ElementType t = new ElementType(s);
    if (view.get(name) != null) {
      if (t.getType() != ElementType.TYPE_TEXT &&
	  t.getType() != ElementType.TYPE_EMPTY)
	throw new IllegalStateException("duplicate value"+t);
      view.put(name, view.get(name)+" "+t.textValue());
    }
    else {
      view.put(name, t.textValue());
    }
  }

  abstract public HashMap getView();

}
 
